# jwt_full/claims.nim
# Copyright 2021 Jonathan Bernard

## ====================================
##               claims
## ====================================
##
## ------------------------------------
##          JWT Claims Object
## ------------------------------------
##
## Data structures and utility procedures for working with JWT claims objects.
## This module follows the common pattern in this library of providing direct
## access to all of the claims members via the `[]` syntax, as well as
## convencience accessors for the publicly-defined claims.

import std/json, std/options, std/times

import ./private/encoding
import ./private/jsonutils
import ./private/timeutils

type
  JwtClaims* = object
    rawB64: string
    json: JsonNode
    iss: Option[string]
    sub: Option[string]
    aud: Option[string]
    exp: Option[DateTime]
    nbf: Option[DateTime]
    iat: Option[DateTime]
    jti: Option[string]

# Public read-only accessors to JwtClaims members
# -----------------------------------------------

func rawB64*(c: JwtClaims): string = c.rawB64
func iss*(c: JwtClaims): Option[string] = c.iss
func sub*(c: JwtClaims): Option[string] = c.sub
func aud*(c: JwtClaims): Option[string] = c.aud
func exp*(c: JwtClaims): Option[DateTime] = c.exp
func nbf*(c: JwtClaims): Option[DateTime] = c.nbf
func iat*(c: JwtClaims): Option[DateTime] = c.iat
func jti*(c: JwtClaims): Option[string] = c.jti

func `[]`*(c: JwtClaims, key: string): Option[JsonNode] =
  ## Generic accessor to claim values by name
  if c.json.hasKey(key): some(c.json[key])
  else: none[JsonNode]()

proc `$`*(claims: JwtClaims): string =
  ## Serialize a JwtClaims object to a Base64url-encoded string (as per the JWT
  ## standard).
  claims.rawB64

proc initJwtClaims*(n: JsonNode): JwtClaims =
  ## Create a JwtClaims from a given set of claims as a JsonNOde

  return JwtClaims(
    rawB64: $n,
    json: n,
    iss: n.optStrVal("iss"),
    sub: n.optStrVal("sub"),
    aud: n.optStrVal("aud"),
    exp: n.optNumericDate("exp"),
    nbf: n.optNumericDate("nbf"),
    iat: n.optNumericDate("iat"),
    jti: n.optStrVal("jtu"))

proc initJwtClaims*(encoded: string): JwtClaims =
  ## Parse a Base64url-encoded set of claims into a JwtClaims object

  let decoded = b64UrlDecode(encoded)
  result = initJwtClaims(parseJson(decoded))
  result.rawB64 = encoded

proc initJwtClaims*(
    iss: Option[string] = none[string](),
    sub: Option[string] = none[string](),
    aud: Option[string] = none[string](),
    exp: Option[DateTime] = none[DateTime](),
    nbf: Option[DateTime] = none[DateTime](),
    iat: Option[DateTime] = none[DateTime](),
    jti: Option[string] = none[string](),
    moreClaims: seq[tuple[k, v: string]] = @[]
  ): JwtClaims =
  ## Convenience method to initialize a new JwtClaims object by specifying
  ## fields directly.
  runnableExamples:
    let claims = initJwtClaims(
      iss = "https://issuer-id",
      sub = "jqt",
      iat = now(),
      exp = now() + 10.minutes,
      moreClaims: @[("name", "John Q. Tester")])

  let json = newJObject()
  for claim in moreClaims: json[claim.k] = %claim.v

  if iss.isSome: json["iss"] = %iss.get
  if sub.isSome: json["sub"] = %sub.get
  if aud.isSome: json["aud"] = %aud.get
  if exp.isSome: json["exp"] = %toNumericDate(exp.get)
  if nbf.isSome: json["nbf"] = %toNumericDate(nbf.get)
  if iat.isSome: json["iat"] = %toNumericDate(iat.get)
  if jti.isSome: json["jti"] = %jti.get

  return initJwtClaims(json)
